#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include <string.h>
#include <malloc.h>
#include "SDL_Scope.h"
#include "AudioFilter.h"
#include "SDL_gfxPrimitives.h"

#define _FXP_X_PER_TICK (TO_FXP(__scope->w)/1000)
#define FRAME_RATE 5
#define FRAME_TICK (TICKS_PER_SECOND/FRAME_RATE)

#define SAMPLE_SIZE 2

#define INITIAL_BORDER 1
int __border = INITIAL_BORDER;

//Color of the oscilloscope in RGBA
#define SCOPE_COLOR		65,255,196,255
#define GRID_COLOR		0,0,255,128
#define ORIGIN_COLOR	65,255,255,196

#define FINE_GRID		32	//# of fine gridlines
#define ROUGH_GRID		8	//# of rough gridlines

//Resolution used for fullscreen
#define FULLSCREEN_W	640
#define FULLSCREEN_H	480

//Macro to resize the window to get properly-spaced gridlines
#define WINDOW_RES(x)	(((x)-((x)%FINE_GRID)+(__border*2)))

#ifdef _WIN32_WCE
#	define W_SIZE		(240-16)
#else
#	define W_SIZE		256	//w/h the window defaults to
#endif//_WIN32_WCE

#define MIN_SIZE		128	//minimum allowable w/h

#define SKIP_LEVEL		4	//Draws one out of every SKIP_LEVEL samples.

#define EVENT_TICKS		10	//How many ticks it idles in the event loop

//Return values for main()
#define MAIN_ERROR -1
#define MAIN_OK		0

//Global variables
SDL_Surface *__screen;		//Global video surface
SDL_Surface *__scope;		//Surface for the 'scope' background
volatile int __running=0;	//Whether the event loop is running or not
SDL_mutex *__lock = NULL;	//Mutex to control access to screen
int fxp_x_per_tick=0;

Sint32 Range(Sint32 max, Sint32 min, Sint16 val);

void DrawGrid(SDL_Surface *surf);


int main(int argc, char *argv[])
{
	SDL_Event event;
	SDL_AudioSpec expected,result;

	int prevw,prevh;
	int fullscreen=0;

	AudioFilterInit();

	//Initialize SDL.
	if(SDL_Init(SDL_INIT_VIDEO)<0)
	{
		fprintf(stderr,"Can't init SDL!\n");
		return(MAIN_ERROR);
	}

	//Open a window, 16bpp preferred but we'll take anything
	__screen=SDL_SetVideoMode(W_SIZE-(W_SIZE%FINE_GRID)+__border*2,W_SIZE-(W_SIZE%FINE_GRID)+__border*2,16,SDL_ANYFORMAT|SDL_RESIZABLE);

	__scope=SDL_DisplayFormat(__screen);
	if(__scope==NULL)
	{
		fprintf(stderr,"Couldn't create backbuffer..\n");
		SDL_Quit();
		return(MAIN_ERROR);
	}
	else
	{
		DrawGrid(__scope);
	}

	//If a window didn't get opened, print error and quit
	if(__screen==NULL)
	{
		fprintf(stderr,"Can't init video!\n");
		SDL_Quit();
		return(MAIN_ERROR);
	}

	SDL_BlitSurface(__scope,NULL,__screen,NULL);

	__lock = SDL_CreateMutex();

	expected.format=AUDIO_FORMAT;
	expected.freq=AUDIO_SAMPLERATE;
	expected.callback=SoundInCallback;
	expected.samples=AUDIO_SAMPLES;
	expected.channels=1;
	expected.userdata=&result;

	if(SDL_OpenAudioIn(&expected,&result)<0)
	{
		fprintf(stderr,"Can't open audio input!\n");
		SDL_Quit();
		return(MAIN_ERROR);
	}

	__running=1;

	fxp_x_per_tick=_FXP_X_PER_TICK;

	//Start the audio
	SDL_PauseAudioIn(0);



	while(__running)
	{
		//Poll for events - keyboard, quit, resize
		while(SDL_PollEvent(&event))
		switch(event.type)
		{
			SDLMod keymod;

			case SDL_QUIT:
				fprintf(stderr,"Quit message recieved, quitting\n");
				__running=0;
				break;

			case SDL_KEYDOWN:
				keymod=SDL_GetModState();
				switch(event.key.keysym.sym)
				{
				case SDLK_KP4:	/* Decrease scroll speed */
				case SDLK_LEFT:
					fxp_x_per_tick*=0.9f;
					break;

				case SDLK_KP6:	/* Increase scroll speed */
				case SDLK_RIGHT:
					fxp_x_per_tick*=(1.0f/0.9f);
					break;

				case SDLK_ESCAPE:
					fprintf(stderr,"ESC pressed, quitting\n");
					__running=0;
					break;

				case SDLK_RETURN:	/* Toggle Fullscreen */
				case SDLK_KP_ENTER:				
					if(keymod&(KMOD_LALT|KMOD_RALT))
					{
						SDL_mutexP(__lock);
						if(fullscreen)
						{							
							fullscreen=0;

							__border=1;

							__screen=SDL_SetVideoMode(prevw,prevh,16,SDL_ANYFORMAT|SDL_RESIZABLE);
							SDL_FreeSurface(__scope);
							if(__screen==NULL)
							{
								fprintf(stderr,"Couldn't change from fullscreen to windowed!\n");
								SDL_CloseAudioIn();								
								SDL_DestroyMutex(__lock);
								SDL_Quit();
								return(MAIN_ERROR);
							}
							else
							{
								__scope=SDL_DisplayFormat(__screen);
								DrawGrid(__scope);
								SDL_ShowCursor(SDL_ENABLE);
							}
						}
						else
						{
							fullscreen=1;

							__border=16;

							prevw=__screen->w;
							prevh=__screen->h;

							__screen=SDL_SetVideoMode(FULLSCREEN_W,FULLSCREEN_H,16,SDL_ANYFORMAT|SDL_FULLSCREEN);
							SDL_FreeSurface(__scope);
							if(__screen==NULL)
							{
								fprintf(stderr,"Couldn't change from windowed to fullscreen!\n");
								SDL_CloseAudioIn();								
								SDL_DestroyMutex(__lock);
								SDL_Quit();
								return(MAIN_ERROR);
							}
							else
							{								
								__scope=SDL_DisplayFormat(__screen);
								DrawGrid(__scope);
								SDL_ShowCursor(SDL_DISABLE);
							}
						}

						SDL_BlitSurface(__scope,NULL,__screen,NULL);
						SDL_Flip(__screen);
						fxp_x_per_tick=_FXP_X_PER_TICK;
						SDL_mutexV(__lock);
					}
				}
				break;

			case SDL_VIDEORESIZE: /* Window resize event */
			{
				int w=event.resize.w,h=event.resize.h,s;

				if(w<h)	s=w;
				else	s=h;

				if(s<MIN_SIZE)	s=MIN_SIZE;

				SDL_mutexP(__lock);
				
				__screen=SDL_SetVideoMode(WINDOW_RES(s),WINDOW_RES(s),16,SDL_ANYFORMAT|SDL_RESIZABLE);

				SDL_FreeSurface(__scope);
				__scope=SDL_DisplayFormat(__screen);
				DrawGrid(__scope);
				SDL_BlitSurface(__scope,NULL,__screen,NULL);
				fxp_x_per_tick=_FXP_X_PER_TICK;
				SDL_mutexV(__lock);				
			}
			break;
		}
		SDL_Delay(EVENT_TICKS);	/* Wait 'till next frame */
	}

	SDL_mutexP(__lock);
	
	SDL_CloseAudioIn();
	SDL_FreeSurface(__scope);
	SDL_DestroyMutex(__lock);
	SDL_Quit();

	return MAIN_OK;
}

void SoundInCallback(void *udata, Uint8 *stream, int len)
{
	int i=0;
	SDL_AudioSpec *spec=(SDL_AudioSpec *)udata;
	Sint16 *sData=(Sint16 *)stream;
	static FXP xPos=TO_FXP(INITIAL_BORDER);
	static FXP xPrev=TO_FXP(INITIAL_BORDER);
	static Uint32 yPrev=(W_SIZE)/2;
	int samples=(len/SAMPLE_SIZE)-1;

	FXP dx=FXP_MUL(fxp_x_per_tick,FXP_TICKS_PER_SAMPLE);
	FXP NewXpos=xPos+(dx*samples);

	AudioFilter(sData,len/2);

	if(__running)
	{
		SDL_Rect RedrawRect;
		SDL_mutexP(__lock);

		RedrawRect.h=__screen->h;
		RedrawRect.x=FROM_FXP(xPos);
		RedrawRect.y=0;
		RedrawRect.w=FROM_FXP(NewXpos-xPos)+1;

		if(SDL_MUSTLOCK(__screen))
		{
			printf("<");
			SDL_LockSurface(__screen);
		}

		while((xPos<NewXpos)&&(i+SKIP_LEVEL<samples))
		{
			xPrev=xPos;
			xPos+=dx*SKIP_LEVEL;

			if(xPos>TO_FXP(__screen->w-__border))
			{
				NewXpos-=TO_FXP(__screen->w);
				xPos=TO_FXP(__border);
				xPrev=TO_FXP(__border);
				SDL_BlitSurface(__scope,NULL,__screen,NULL);
				RedrawRect.x=0;
				RedrawRect.w=__screen->w;
			}

			aalineRGBA(__screen,FROM_FXP(xPrev),yPrev,FROM_FXP(xPos),Range(__screen->h,0,sData[i]),SCOPE_COLOR);
			yPrev=Range(__screen->h,0,sData[i]);
			i+=SKIP_LEVEL;
		}

		if(SDL_MUSTLOCK(__screen))
		{
			printf(">");
			SDL_UnlockSurface(__screen);
		}


		SDL_UpdateRect(__screen,RedrawRect.x,RedrawRect.y,RedrawRect.w,RedrawRect.h);

		SDL_mutexV(__lock);
	}
}

Sint32 Range(Sint32 max, Sint32 min, Sint16 val)
{
	Sint32 tVal=((Sint32)val)+32768;
	tVal*=(max-min);
	tVal/=65536;
	return((Sint32)tVal);
}

void DrawGrid(SDL_Surface *surf)
{
	int x,y;

	if(surf==NULL) return;

	for(x=__border; x<=surf->w-__border; x+=((surf->w-(2*__border))/FINE_GRID))
	{	vlineRGBA(surf,x,__border,surf->h-__border,GRID_COLOR);	}

	for(y=__border; y<=surf->h-__border; y+=((surf->h-(2*__border))/FINE_GRID))
	{	hlineRGBA(surf,__border,surf->w-__border,y,GRID_COLOR);	}

	for(x=__border; x<=surf->w-__border; x+=((surf->w-(2*__border))/ROUGH_GRID))
	{	vlineRGBA(surf,x,__border,surf->h-__border,GRID_COLOR);	}

	for(y=__border; y<=surf->h-__border; y+=((surf->h-(2*__border))/ROUGH_GRID))
	{	hlineRGBA(surf,__border,surf->w-__border,y,GRID_COLOR);	}

	vlineRGBA(surf,surf->w/2,__border+1,surf->h-__border-1,ORIGIN_COLOR);
	hlineRGBA(surf,__border+1,surf->w-__border-1,surf->h/2,ORIGIN_COLOR);
}
